#include <math.h>
#include <string.h>
#include <float.h>
#include <stdlib.h>

#include <iostream>
#include <fstream>
#include <vector>
#include <string>

#include "rts_feature.hpp"
#include "rts_sample.hpp"
#include "rts_node.hpp"
#include "rts_tree.hpp"
#include "common.h"
#include "includeCV.h"

namespace RTs{

/*!
@brief ċA Node ̊Kw𐶐֐D
@param[in] samples TvW
@return vector<double> AW
*/
std::vector<double> Tree::multipleRegression(const std::vector<const Sample *> &samples){
	cv::Mat vari_cov(INPUT_DIM_SIZE, INPUT_DIM_SIZE, CV_64F, cv::Scalar(0.0));	//	UUs
	std::vector<double> covariance(INPUT_DIM_SIZE, 0.0);	// U
	std::vector<double> R_matrix(INPUT_DIM_SIZE+1, 0.0);	// AW
	std::vector<double> input_ave(INPUT_DIM_SIZE, 0.0);		// ͕ϐ̕
	double output_ave = 0.0;								// o͕ϐ̕
	
	//
	//	ŏ@ɂdA
	//
	
	for(unsigned int s=0; s<samples.size(); ++s){
		for(int fx=0; fx<INPUT_DIM_SIZE; fx++)
			// ͕ϐ̑aZo
			input_ave[fx] += samples[s]->feature_vec[fx];

			// o͕ϐ̑aZo
			output_ave += samples[s]->label_vec;

	}
	//	ς̎Zo
	for(int fx=0; fx<INPUT_DIM_SIZE; fx++)
		input_ave[fx] /= samples.size();

		output_ave /= samples.size();

	//UCU̎Zo
	for(int r=0; r<INPUT_DIM_SIZE; r++){
		for(int c=0; c<INPUT_DIM_SIZE; c++){
			double tmp_vari_cov = 0.0;
			for(unsigned int s=0; s<samples.size(); s++){
				tmp_vari_cov += (input_ave[r] - samples[s]->feature_vec[r]) * (input_ave[c] - samples[s]->feature_vec[c]);
			}
			vari_cov.at<double>(r, c) = tmp_vari_cov / (double)samples.size();
		}
		//	o͕ϐƂ̋U
			for(unsigned int s=0; s<samples.size(); s++){
				covariance[r] += (input_ave[r] - samples[s]->feature_vec[r]) * (output_ave - samples[s]->label_vec);
			}
			covariance[r] /= samples.size(); 
	}
	//std::cout << "vari_cov = " << vari_cov << std::endl;

	//	UUs̋tsZo
	vari_cov = vari_cov.inv(cv::DECOMP_SVD);

	//std::cout << "ts = " << vari_cov << std::endl;
		
	//	AW̎Zo
	for(int r=0; r<INPUT_DIM_SIZE; r++){
		double tmp_var = 0.0;
		for(int c=0; c<INPUT_DIM_SIZE; c++){
			tmp_var += vari_cov.at<double>(r, c) * covariance[c];
		}
		R_matrix[r] = tmp_var; 
	}

	//	AW̒萔Zo
	double tmp_r = 0.0;
	for(int fx=0; fx<INPUT_DIM_SIZE; fx++)
		tmp_r += R_matrix[fx] * input_ave[fx];
	
	R_matrix[INPUT_DIM_SIZE] = output_ave - tmp_r;
	
	//	ʂ̏o
	//std::cout << "y = ";
	//for(int fx=0; fx<INPUT_DIM_SIZE; fx++)
	//std::cout << R_matrix[fx] << "*x" << fx << " + ";

	//std::cout << R_matrix[INPUT_DIM_SIZE] << std::endl;

	return R_matrix;
}

/*!
@brief ċA Node ̊Kw𐶐֐D
@param[in] maxDepth ؂̊Kw̍ő吔
@param[in] featureTrials ʃ_I̎s
@param[in] thresholdTrials l_I̎s
@param[in] samples tM
@param[in] inverse_label_freq x̕px̋t (dݕtɎg)
@param[in] depth ؂̓ɍ琶 Node ̐[
@return node
*/

Node *Tree::build(const int maxDepth,
                  const int featureTrials,
                  const int thresholdTrials,
                  const std::vector<const Sample *> &samples,
                  const int depth){

	if(samples.empty()){
		return NULL;
	}

	// ʂ̎Dƃ_TCTv̓ʂ̎QƂD
	unsigned int feature_dim = samples[0]->feature_vec.size();

//	std::cout << feature_dim << std::endl;

	// ̂l_œʂ̊e̍őlCŏlKvȂ̂ŁCŐɋ߂Ă
	std::vector<float> feature_min_values(feature_dim);
	std::vector<float> feature_max_values(feature_dim);
	for(unsigned int f=0; f<feature_dim; ++f){
		feature_min_values[f] = feature_max_values[f] = samples[0]->feature_vec[f];
	}
	for(unsigned int s=1; s<samples.size(); ++s){
		const Feature *feature_vec = &(samples[s]->feature_vec);
		if(feature_vec->size() != feature_dim){
			// G[Fʂ̎ĂȂ??
			return NULL;
		}
		for(unsigned int f=0; f<feature_dim; ++f){
			float feature_value = (*feature_vec)[f];
			if(feature_value < feature_min_values[f]){
				feature_min_values[f] = feature_value;
			}
			else if(feature_max_values[f] < feature_value){
				feature_max_values[f] = feature_value;
			}
		}
	}

	// ʂ̕ێpobt@
	std::vector<const Sample *> lSamples;
	std::vector<const Sample *> rSamples;
	// KvɂȂ蓾TCY̍ől̃obt@mۂĂ
	lSamples.reserve(samples.size());
	rSamples.reserve(samples.size());

	// Uŏ镪ʂT
	double bestGain = DBL_MAX;
	float bestThreshold = 0;
	int bestFeature = 0;
	std::vector<const Sample *> bestlSamples;
	std::vector<const Sample *> bestrSamples;
	bestlSamples.reserve(samples.size());
	bestrSamples.reserve(samples.size());

	for(int ft=0; ft<featureTrials; ++ft){
		
		// ɂʂ̎_Ɍ
		int rand_feature_id = (int)(rand() % feature_dim);
		
		for(int tt=0; tt<thresholdTrials; ++tt){
			
			// ɂl "rand_feature_id" ̍őEŏ̊ԂŃ_Ɍ
			float rand_threshold = feature_min_values[rand_feature_id]
			                      +(float)rand()/(float)(RAND_MAX)
			                       *(feature_max_values[rand_feature_id] - feature_min_values[rand_feature_id]);

			//
			// L̃_Ȃlŕ
			//

			lSamples.clear();
			rSamples.clear();

			//Tv𕪊
			for(unsigned int s=0; s<samples.size(); ++s){
				if(samples[s]->feature_vec[rand_feature_id] < rand_threshold){
					//	qm[h̃Tvɒǉ
					lSamples.push_back(samples[s]);
				}else{
					//	E̎qm[hɃTvǉ
					rSamples.push_back(samples[s]);
				}
			}

			//̎qm[h̉ǍWZo@
			std::vector<double> rMatrix_left = multipleRegression(lSamples);
			//E̎qm[h̉ǍWZo
			std::vector<double> rMatrix_right = multipleRegression(rSamples);

			double left_gain = 0.0;
			double right_gain = 0.0;
			
			//̎c̎Zo
			for(unsigned int left=0; left<lSamples.size(); left++){
				double tmp_left = 0.0;
				for(unsigned int fx=0; fx<lSamples[left]->feature_vec.size(); fx++){
					tmp_left += rMatrix_left[fx] * (double)lSamples[left]->feature_vec[fx]; 
				}
				tmp_left += rMatrix_left[lSamples[left]->feature_vec.size()];
				left_gain += (tmp_left - lSamples[left]->label_vec)*(tmp_left - lSamples[left]->label_vec);
			}
			//E̎c̎Zo
			for(unsigned int right=0; right<rSamples.size(); right++){
				double tmp_right = 0.0;
				for(unsigned int fx=0; fx<rSamples[right]->feature_vec.size(); fx++){
					tmp_right += rMatrix_right[fx] * (double)rSamples[right]->feature_vec[fx]; 
				}
				tmp_right += rMatrix_right[rSamples[right]->feature_vec.size()];
				right_gain += (tmp_right - rSamples[right]->label_vec)*(tmp_right - rSamples[right]->label_vec);
			}

			//
			// c̍ŏlT
			//

			double gain = lSamples.size() * left_gain + rSamples.size() * right_gain;

			if(gain < bestGain){
				bestGain = gain;
				bestFeature = rand_feature_id;
				bestThreshold = rand_threshold;
				bestlSamples = lSamples;
				bestrSamples = rSamples;
			}
		

		//std::cout << "œKȕ֐TĂ܂"
		  //        << " [" << ft+1 << "/" << featureTrials << "]" << "gain = " << bestGain << "\r";
//std::cout.flush();
		}
	}
	//std::cout << std::endl;

	// ŗǂp[^m[hɕۑ
	Node *node = new Node();
	node->feature_id = bestFeature;
	node->threshold = bestThreshold;

	//
	// Ŗ[m[hɓBǂ`FbN
	//
	if(FLT_MIN > bestGain || maxDepth < depth || bestlSamples.empty() || bestrSamples.empty() || samples.size() < MIN_SAMPLE_SIZE){
		node->flag = false;
		//std::cout << "leaf node !" << std::endl;
	}
	else{
		node->flag = true;
		// ܂m[hȂ̂ŁCqm[h𐶐
		node->lChild = build(maxDepth, featureTrials, thresholdTrials, bestlSamples, depth+1);
		node->rChild = build(maxDepth, featureTrials, thresholdTrials, bestrSamples,  depth+1);
		if(node->lChild == NULL || node->rChild == NULL){
			// qm[h̐ߒŉG[N͗l
			delete node;
			node = NULL;
		}
	}

	return node;

}


/*!
@brief tM𖖒[m[h̃[tɓ[֐D
@param[in] sample 瓊[鋳tM
@param[in] node `FbN Node
@return succeed or not
*/

bool Tree::voteSample(const Sample* &sample,
                          Node *node){
	if(node == NULL){
		return false;
	}

	if(node->flag == true){
		// ̃m[h͕m[hȂ̂ŁC؂̎qm[h֐i
		if(sample->feature_vec[node->feature_id] < node->threshold){
			return voteSample(sample, node->lChild);
		}
		else{
			return voteSample(sample, node->rChild);
		}
	}
	else{
		// ̃m[h͖[m[hȂ̂ŗ\ɉZ
		node->samples.push_back(sample);
	}

	return true;

}


/*!
@brief ċAŖ[m[hTC[m[h̗\쐬֐D
@param[in] node `FbN Node
@return succeed or not
*/

bool Tree::normalizeLeaf(Node *node){

	if(node == NULL){
		return false;
	}

	if(node->flag == true){
		// ̃m[h͕m[hȂ̂ŁC؂̎qm[h֐i
		if(normalizeLeaf(node->lChild) == false
		|| normalizeLeaf(node->rChild) == false){
			return false;
		}
	}
	else{
		// ̃m[h͖[m[hȂ̂ōŏ@ɂ`dA
		node->coefficient = multipleRegression(node->samples);
	}

	return true;

}


/*!
@brief ċAŖ[m[hTC[m[h̗\Ԃ֐D
@param[in] node `FbN Node
@param[in] feature_vec ʑΏۂ̓
@return histogram
*/

double Tree::traversal(Node *node,
                             const Feature &feature_vec){

	if(node == NULL){
		return NULL;
	}

	if(node->flag == true){
		// ̃m[h͕m[hȂ̂ŁC؂̎qm[h֐i
		if(feature_vec[node->feature_id] < node->threshold){
			return traversal(node->lChild, feature_vec);
		}
		else{
			return traversal(node->rChild, feature_vec);
		}
	}

	// ̃m[h͖[m[hȂ̂ŗ\Ԃ
	double prediction = 0.0;
	for(int fx=0; fx<INPUT_DIM_SIZE; fx++){
		prediction += feature_vec[fx] * node->coefficient[fx];
	}
	prediction += node->coefficient[INPUT_DIM_SIZE];
	return prediction;
}


/*!
@brief ċAŌ؍\̑SẴm[hۑ֐D
@param[in] ofs o̓Xg[
@param[in] node ۑ Node
@return succeed or not
*/

bool Tree::save(std::ofstream &ofs,
                Node *node){

	if(node == NULL){
		return false;
	}

	// ̃m[h̏
	ofs << node->feature_id << " "
		<< node->threshold << " "
		<< node->flag << " "
		<< std::endl;
	if(ofs.fail()){
		return false;
	}

	if(node->flag == true){
		// m[hȂ̂ōċAŎqm[h
		if(save(ofs, node->lChild) == false
		|| save(ofs, node->rChild) == false){
			return false;
		}
	}
	else{

		// [m[hȂ̂ŉAW
		for(int fx=0; fx<node->coefficient.size(); fx++){
			ofs << node->coefficient[fx] << " ";
		}
		ofs << std::endl;
		if(ofs.fail()){
			return false;
		}
	}

	return true;

}


/*!
@brief ċAŌ؍\̑SẴm[hǂ݂ފ֐D
@param[in] ifs ̓Xg[
@return node
*/

Node *Tree::load(std::ifstream &ifs){

	Node *node = new Node();

	ifs >> node->feature_id
	    >> node->threshold
		>> node->flag;
	if(ifs.fail()){
		std::cout << "eLXg܂ǂ߂܂ł\n";
		delete node;
		return NULL;
	}

	if(node->flag == false){
		// ̃m[h͖[m[hȂ̂ŉAWǂ݂
		node->coefficient.resize(INPUT_DIM_SIZE+1);
		for(int fx=0; fx<=INPUT_DIM_SIZE; fx++){
			ifs >> node->coefficient[fx];
		}
	}
	else{
		// ̃m[h͕m[hȂ̂Ŏqm[hǂ݂
		node->lChild = load(ifs);
		node->rChild = load(ifs);
		if(node->lChild == NULL || node->rChild == NULL){
			std::cout << "qm[h܂\n";
			delete node;
			return NULL;
		}
	}

	return node;

}


/*!
@brief RXgN^
@param[in] _numClass ʂNX
*/

Tree::Tree(){
	root = NULL;
}


/*!
@brief fXgN^
*/

Tree::~Tree(){
	if(root != NULL){
		delete root;
	}
}


/*!
@brief w肳ꂽp[^ŊwKf[^猈؍\𐶐֐D
@param[in] maxDepth ؂̊Kw̍ő吔
@param[in] featureTrials ʃ_I̎s
@param[in] thresholdTrials l_I̎s
@param[in] samples tM
@param[in] inverse_label_freq x̕px̋t (dݕtɎg)
*/

bool Tree::BuildTree(const int maxDepth,
                     const int featureTrials,
                     const int thresholdTrials,
                      const std::vector<const Sample *> &samples){

	if(root != NULL){
		delete root;
	}
	root = build(maxDepth, featureTrials, thresholdTrials, samples, 0);

	if(root == NULL){
		// G[Nꍇ NULL ԂĂ
		return false;
	}

	return true;

}


/*!
@brief \zς݂̌؍\gCwKf[^Ŗ[m[h̗\𐶐֐D
@param[in] samples 瓊[鋳tM
@param[in] node `FbN Node
@return succeed or not
*/

bool Tree::BuildLeaf( std::vector<const Sample *> &samples){

	if(root == NULL){
		return false;
	}

	// wKf[^PÂ[m[hɓ[
	for(unsigned int s=0; s<samples.size(); ++s){
		if(voteSample(samples[s], root) == false){
			return false;
		}
	}

	// SĂ̖[m[h̃qXgO𐳋K
	if(normalizeLeaf(root) == false){
		return false;
	}

	return true;

}


/*!
@brief ʑΏۂ̓xNgCΉqXgO߂֐D
@param[in] feature_vec xNg
@return qXgO (G[ɂ NULL Ԃ)
*/

double Tree::Traversal(const Feature &feature_vec){
	
	if(root == NULL){
		return NULL;
	}

	return traversal(root, feature_vec);

}


/*!
@brief ؍\ۑ֐D
@param[in] ofs o̓Xg[
@return succeed or not
*/

bool Tree::Save(std::ofstream &ofs){

	if(root == NULL){
		return false;
	}

	return save(ofs, root);

}


/*!
@brief ؍\ǂ݂ފ֐D
@param[in] ifs ̓Xg[
@return succeed or not
*/

bool Tree::Load(std::ifstream &ifs){

	if(root == NULL){
		delete root;
	}

	root = load(ifs);
	if(root == NULL){
		return false;
	}

	return true;

}


} // namespace RTs
